\ 
\ Last_change: KS 15.01.2012 15:59:44
\ 
Forth definitions

include oop.fs

\ -----------------------------------------------------------------------------
\ Basic Objects: Cell, Buffer, String
\ -----------------------------------------------------------------------------

Class Cell   Cell definitions

1 cells Self allot

: @    ( obj -- n )   @ ;
                          
: !    ( n obj -- )   ! ;
                          
: +!   ( n obj -- )   +! ;
                          
: on   ( obj -- )     on ;
                          
: off  ( obj -- )     off ;

: ?    ( obj -- )     @ . ;

Forth definitions

Class Buffer   Buffer definitions

Cell Attribute Ptr
Cell Attribute Len

: size   ( obj -- +n )         Self Len @ ;

: addr   ( obj -- buf.addr )   Self Ptr @ ?dup 0= ABORT" uninitialized_buffer" ;
 
: allot  ( +n obj -- )
   dup Self Ptr @ IF  Self Len @ u> ABORT" buffer_too_small"  EXIT THEN
   here over Self Ptr !   over swap Self Len !   allot
;
: update ( addr +n obj -- )    dup >r Self Len !  r> Self Ptr ! ;

: reset  ( obj -- )            >r 0 0 r> Self update ;

: ?      ( obj -- )            dup @ ." addr " u.  cell+ @ ." length " u. ;

Forth definitions

class String   String definitions

Buffer   Attribute String

Polymorphic @
Polymorphic ??

: @     ( obj -- c-addr +n )   Self String addr count ;

: size  ( obj -- +n )          Self String size 1- ;

: ?size ( +n obj -- +n )
   Self size over < ABORT" string_overflow" ;

: chars ( obj -- +n )          Self @ nip ;

: !     ( c-addr +n obj -- )   >r r@ Self ?size r> Self String addr place ;

: ?     ( obj -- )             Self @ type ;

: init  ( +n obj -- )
    >r &255 umin 1+ r@ Self String allot
    r> Self String addr off
;
: +!    ( c-addr +n obj -- )
    >r r@ Self chars  2dup +  r@ Self ?size  ( addr +n chars newsize )
    dup >r   swap - umin                     ( addr +n' )
    r>   r@ Self @ +                         ( addr +n' newsize addr+ )
    swap r> Self String addr c! \ set new length ( addr +n' addr+ )
    swap move                   \ append string
;

Forth definitions
\ -----------------------------------------------------------------------------
\ testing
\ -----------------------------------------------------------------------------

String Object s1   20 s1 init
String Object s2   20 s2 init
Create dies ," drhn"  dies count s1 !
Create das  ," blah"    das count s2 !

: stringtest  ( -- )  s1 @ s2 +!   s2 ? ;

\ -----------------------------------------------------------------------------
\ Proxy with lazy initialisation
\ -----------------------------------------------------------------------------

Class Indirect  Indirect definitions

Cell Attribute Zahl
Cell Proxy Ref

Forth definitions

Class Refref Refref definitions

Indirect Attribute Indir

Forth definitions

Cell     Object Zelle
Indirect Object Ind
Refref   Object Deeper

: deep-assign ( xt -- )  BIND Deeper Indir Ref ;

: init  ( -- object )   ." loading " ['] Zelle BIND Ind Ref   Ind Ref .. ;
' init BIND Ind Ref

\ -----------------------------------------------------------------------------
\ late binding / forward referencing
\ -----------------------------------------------------------------------------
Forth definitions

: with-dummy ( -- )   s1 ?? ;

\ Later, ?? can be defined as meaningful string method, e.g.:
\ String definitions
\ >: ??  dup Self ? Self ? ;

\ -----------------------------------------------------------------------------
\ polymorphism
\ -----------------------------------------------------------------------------

: set@  ( xt -- )   bind s1 @ ."  polymorphism" ;

:noname  ." poly " string string addr count ; BIND s1 @

