\
\ File name: OOP.FS loads on top of gforth
\ 
\ Last_change: KS 16.01.2012 18:49:13
\
\ with polymorphism on demand

Only Forth also definitions   warnings off  hex

cr .( klausOOP V_2.0 with polymorphism, dynamic late binding, single inheritance, and Proxies.)
cr

\ -----------------------------------------------------------------------------
\ gforth extensions
\ -----------------------------------------------------------------------------

true Constant gforth_062 \ gforth_070 when false

: cell-   1 cells - ;
: ?comp   state @ 0= Abort" compilation only" ;
: ?exec   state @    Abort" execution only" ;

: case?   ( n1 n2 -- n1 ff | tf )   over = dup IF  nip  THEN ;

Create restore ] r> r> ! EXIT [     \ poor man's local variable mechanism
: save  ( var -- )  r> swap dup >r @ >r restore >r >r ;

: becomes  ( xt <name> -- ) \ makes existing <name> behave as xt
   here >r
   >r   ' >body dp !
   postpone ahead   r> >body dp !   postpone THEN
   r> dp !
;
: new.voc ( wid -- )
    dup >r wordlist-struct %size + dup head? -1 = IF ( wid nt )
       dup name>int dup >code-address docon: = swap >body @ r@ = and IF
          id. rdrop EXIT
    THEN THEN
    drop r> body> >head-noprim id. ;
' new.voc becomes .voc

: \\ ( -- )
   SOURCE-ID  1+ 1 U> IF  BEGIN  REFILL 0=  UNTIL THEN
   POSTPONE \
; IMMEDIATE

: set   ( mask addr -- )   dup >r @ or r> ! ;

: reset ( mask addr -- )   >r invert r@ @ and r> ! ;

\ -----------------------------------------------------------------------------
\ OOP mechanisms on top of gforth
\ -----------------------------------------------------------------------------

Vocabulary Oop   Oop also definitions

Variable (doClass  : doClass ( -- xt )  (doClass @ ;
Variable (doObj    : doObj   ( -- xt )  (doObj @ ;
Variable (doAttr   : doAttr  ( -- xt )  (doAttr @ ;
Variable (doProxy  : doProxy ( -- xt )  (doProxy @ ;
Variable (doPoly   : doPoly  ( -- xt )  (doPoly @ ;

Variable LastClass     LastClass off
Variable ClassContext  ClassContext off    \ Current class context

: Forth-context ( -- )  ClassContext off ;

: set-class   ( class -- )  LastClass !  ClassContext on ;

Variable 'classroot  'classroot off        \ serves as a forward reference

Forth definitions

: Class       ( <name> -- )
   Vocabulary  0 , 0 , 'classroot @ , immediate
\ | search-xt | wordlist | Voclink | ext || attr | size | parent |
DOES>  ( -- )  [ here (doClass ! ] set-class ;

Oop definitions

: classorder  ( -- ) \ establish Forth search inside a :-definition
   r> context save >r
   LastClass @ ?dup 0= Abort" Class context undefined"
   context !
;
: CurrentClass     ( -- cpa )
   Current @ cell- @ doClass - Abort" No class compilation context set"
   Current
;
: c.last   ( cpa -- addr )  @ cell+ ;                   \ anchor of class wordlist

: c.attr   ( cpa -- addr )  @ [ 4 cells ] Literal + ;
   1 Constant #sealed                                   \ attribute bit
   : ?unsealed   ( cpa -- )                             \ all other bits are unused at present
      c.attr @ #sealed and Abort" Class is sealed"
   ;
: seal-class  ( -- )   #sealed LastClass c.attr set ;

: c.size   ( cpa -- addr )  @ [ 5 cells ] Literal + ;

: c.parent ( cpa -- addr )  @ [ 6 cells ] Literal + ;

: search-methods  ( addr len listaddr -- nfa | 0 )
   BEGIN  >r 2dup r@ @ (search-wordlist)
          ?dup IF  nip nip rdrop  EXIT THEN
          r> c.parent   dup @ 0=
   UNTIL
   drop 2drop false
;
: class'    ( <name> -- xt )
   name LastClass search-methods
   dup 0= Abort" ?" name>int
;
: OopCreate ( <name> -- )    Create LastClass @ , immediate ;

: >object   ( addr -- obj )  dup @ set-class   cell+ ;

: :cfa,     ( -- )           docol: , 0 , ;

: ==        ( xt u -- xt f )   over cell+ @ = ;

\ -----------------------------------------------------------------------------
\ outer interpreter for gforth_0.62
\ -----------------------------------------------------------------------------

: oop-compiler ( c-addr u -- )
   ClassContext @   Forth-context
   IF  2dup LastClass search-methods ?dup
      IF  nip nip name>comp execute  EXIT
      THEN
      2drop  compiler-notfound EXIT
   THEN
   2dup find-name ?dup
   IF  nip nip name>comp execute  EXIT
   THEN
   2dup snumber? ?dup
   IF  0> IF  swap postpone Literal  THEN
       postpone Literal  2drop  EXIT
   THEN
   compiler-notfound
;
: oop-interpreter ( c-addr u -- )
   ClassContext @
   IF  2dup LastClass search-methods ?dup
      IF  Forth-context  nip nip name>int execute  EXIT
      THEN
      2drop  interpreter-notfound EXIT
   THEN
   2dup find-name ?dup
   IF  nip nip name>int execute  EXIT
   THEN
   2dup 2>r snumber? IF  2rdrop  EXIT THEN
   2r> interpreter-notfound
;
gforth_062 [IF]

\ these re-definitions cannot be done using >: - I have no idea why

   : new[ ( -- )  ['] oop-interpreter IS parser state off ; immediate
   : new] ( -- )  ['] oop-compiler    IS parser state on  ;
   ' new[ becomes [
   ' new] becomes ]
   [

[ELSE]

cr .( this should hold code for gforth_071 but it doesn't !!! )
\ : host-compiler1  ( addr len -- xt )  host-compiler ['] noop ;
\ ' host-compiler1 IS parser1

[THEN]
\ -----------------------------------------------------------------------------
\ OOP support in the Root wordlist
\ -----------------------------------------------------------------------------
Forth Root definitions Forth

: Self     ( -- )       CurrentClass @ set-class ; immediate

: classes  ( -- )  space
   Voclink  BEGIN  @ ?dup
   WHILE  dup 2 cells -  dup body> cell+ @ doClass =
          IF  dup body> >name .name  THEN  drop
   REPEAT
;
: methods  ( -- )   LastClass
   BEGIN  dup c.parent @
   WHILE  dup @ dup cr .voc ." :   " wordlist-id
          BEGIN  @ ?dup WHILE  dup .name  REPEAT
          c.parent
   REPEAT  drop
;
\ -----------------------------------------------------------------------------
\ Polymorphism on demand
\ -----------------------------------------------------------------------------
\ Polymorphic is a defining word, which reserves one field in each instantiated
\ object that holds the xt of the actual method. It can only be used as long
\ as the class is not sealed.
\ Immediately following the polymorphic data type is a branch to the latest
\ code that will be used to initialize the object's associated xt-field.
\ Initially, this branch point to code that will print an "uninitialized" error
\ message. Later on, a :-definition of the same name will modify the branch,
\ to point to it.
\ See: set-polymorphics in "Object" for initializing the xt
\ -----------------------------------------------------------------------------
Oop definitions

: set-polymorphics  ( obj -- ) \ initialize polymorphic xt's of an object
   >r   LastClass
   BEGIN  dup c.last
      BEGIN  @ ?dup
      WHILE  dup name>int cell+ @ dopoly =
             IF  dup name>int >body
                 dup @ r@ + @ 0=
                 IF  dup cell+   over @ r@ + !  THEN
                 drop
             THEN
      REPEAT
      c.parent
      dup c.parent @ 0=                    \ don't search ClassRoot
   UNTIL
   drop rdrop
;
: do-polymorphic  ( obj offset -- )   over + @ ?dup 0= Abort" un-initialized poly-field in object" execute ;

: un-initialized   true Abort" un-initialized message" ;

Forth definitions

: Polymorphic ( -- )
   CurrentClass ?unsealed
   Create immediate
           CurrentClass c.size @ ,          \ offset into object's data field
   1 cells CurrentClass c.size +!
   :cfa, postpone un-initialized postpone EXIT
\ <name> | offset | :-definition of actual code
DOES> ( obj -- ) [ here (doPoly ! ]
   @
   State @ IF  postpone Literal postpone do-polymorphic  EXIT THEN
   do-polymorphic
;
\ -----------------------------------------------------------------------------
\ ClassRoot and initial methods
\ -----------------------------------------------------------------------------

Class ClassRoot                \ Note: ClassRoot's parent field is set to 0
' ClassRoot >body 'classroot ! \ every class inherits ClassRoot
' ClassRoot >body set-current  ( ClassRoot definitions )

: definitions  ( -- )    LastClass @ Current ! ;

: '       ( <name> -- )   class' ;

: Class   ( <name> -- )  Class
   LastClass c.size @ here 2 cells - !                  \ inherits lastclass
   LastClass @ here cell- !   seal-class
;
: Object   ( <name> -- )
   OopCreate
   here   LastClass c.size @ allot
   set-polymorphics   seal-class
DOES> ( -- obj )  [ here (doObj ! ]
   >object  State @ IF  postpone Literal  THEN
;
: Attribute ( <name> -- )
   CurrentClass ?unsealed
   OopCreate          CurrentClass c.size @ ,
   LastClass c.size @ CurrentClass c.size +!            \ make room for included object
DOES>  ( obj1 -- obj2 )  [ here (doAttr ! ]
   >object @    ?dup 0= ?EXIT                           \ don't do anything when the offset is 0
   State @ IF  postpone Literal postpone +  EXIT THEN   \ could be optimized by adding to preceeding Literal
   +
;
Forth definitions

: do-proxy  ( obj offset  -- )  + @ ?dup 0= Abort" un-initialized Proxy" execute ;

ClassRoot definitions

: Proxy ( <name> -- )
   CurrentClass ?unsealed
   OopCreate  CurrentClass c.size @ ,                   \ offset into the object. The cell holds an execution token
   1 cells    CurrentClass c.size +!                    \ room for execution token of objects or initialization
DOES> ( obj1 -- obj2 )  [ here (doProxy ! ]
   >object @
   State @ IF  postpone Literal postpone do-proxy  EXIT THEN  \ could be optimized by adding to preceeding Literal
   do-proxy
;
: allot   ( u -- )        CurrentClass c.size +! ;

: units   ( u1 -- u2 )    LastClass c.size @ * ;

: addr    ( obj -- addr ) Forth-context ; immediate

: ..      ( -- )          Forth-context ; immediate

: size    ( -- bytes )    ?exec LastClass c.size @ ;

: words   ( -- )          classorder words ;

: methods ( -- )          methods ;

: order   ( -- )  LastClass >r
   BEGIN  r@ @ ?dup WHILE  .voc  r> c.parent >r  REPEAT  rdrop
   ."     "  Current @ .voc
;
: see     ( <name> -- )   classorder see  ;

\ -----------------------------------------------------------------------------
\ intelligent : and ; to take care of re-defining methods and messages
\ -----------------------------------------------------------------------------
Oop definitions

0 Constant #colon    \ type digits to
1 Constant #method   \ control ;'s behaviour

: ?polymorphic  ( xt1 -- xt2 )   dopoly == IF  >body cell+  THEN ;

: isBranch?  ( x -- f )                             \ test x to be a valid branch code
   drop true                                        \ unfortunately, I don't know how to do this for gforth
;
: forth-colon?  ( -- f )
   Current @ cell- @ doClass -                      \ not a class definition
   Current @ [ ' Classroot >body ] Literal = or     \ nor ClassRoot definitions
;
Forth definitions

: :  ( <name> -- sys # )
   forth-colon? IF  : #colon  EXIT THEN
   name   2dup Current @ (search-wordlist) ?dup
   IF  nip nip                                      \ re-define an existing, non-polymorphic method
       name>int ?polymorphic >body                  \ leave addr of branch to be rewritten
       align :cfa,   here                           \ leave addr of definitions body
       defstart ] :-hook   #method  EXIT            \ leave check number
   THEN
   2dup CurrentClass search-methods ?dup
   IF  name>int dopoly ==                           \ (re)define a polymorphic method
       IF  >r                                       \ xt of previously defined polymorphic with the same name
           header, dodoes: , dopoly ,  r> >body @ , \ identical index as previous one
           :cfa,
           postpone AHEAD  :cfa,  postpone THEN     \ this :-cf is needed to make "embed" possible
           defstart ] :-hook #colon EXIT
       THEN
       drop
   THEN
   header, (:noname)                                \ a new, non-polymorphic method
   postpone AHEAD  :cfa,  postpone THEN             \ this :-cf is needed to make "embed" possible
   #colon
;
: ;       ( sys # -- )
   #colon  case? IF  postpone ;  EXIT THEN
   #method case? IF  postpone ;                     \ create a branch from the previously
                     here >r                        \ defined non-polymorphic method
                     >r dp !  postpone AHEAD        \ to its new definition
                     r> dp !  postpone THEN
                     r> dp !  EXIT
                 THEN
   Abort" unstructured!"
[ drop ] ; immediate                                \ this ";" for the last time does not use a type digit

: Does>   ( # -- # )     >r postpone Does> r> ; immediate

: :noname ( -- # )       :noname #colon ;

\ -----------------------------------------------------------------------------
\ embed is used to embed the previously valid code of a method into a
\ re-definition of that method
\ -----------------------------------------------------------------------------

: embed   ( <name> -- )   \ compile previous version of <name> into <name>'s redefinition
   postpone Self  class'  ?polymorphic
   >body Forth-context
   dup @ isBranch? 0= Abort" not a method definition"
   cell+ @ body> compile,
; immediate

\ -----------------------------------------------------------------------------
\ bind is used to initialize or modify the xt of a Proxy or Polymorphic method
\ in a concrete object.
\ -----------------------------------------------------------------------------

: bind     ( xt <proxy name> -- )                            \ bind does not type check, so xt can be anything
   State @ >r   State off                                    \ State must be off, because executed objects below are state smart
   class' doObj  == 0= Abort" no object"  execute            \ object containing reference
   BEGIN  dup class' doAttr == WHILE  nip execute  REPEAT    \ skip over potential hierarchy of attributes and sub-classes
   nip   r> State !   Forth-context
   doProxy == IF  >body cell+
   ELSE
      doPoly == 0= Abort" no proxy nor method"  >body
   THEN
   @ +
   State @ IF  postpone Literal postpone !  EXIT THEN
   !
; immediate

Only Forth also definitions
